import errno
import os
import pty
import struct
from contextlib import contextmanager
from dataclasses import dataclass
from multiprocessing.pool import Pool
import signal
import fcntl
from typing import Optional

import termios


def is_session_leader():
    return os.getsid(0) == os.getpid()


def set_group_leader():
    os.setpgid(0, 0)
    return os.getpgid(0)


def make_sure_not_group_leader():
    if os.getpgid(0) == os.getpid():
        with Pool(1) as pool:
            pgid = pool.apply(set_group_leader)
            os.setpgid(0, pgid)


def make_session_leader():
    make_sure_not_group_leader()
    os.setsid()


@contextmanager
def set_handler(sig, handler):
    old = signal.signal(sig, handler)
    try:
        yield old
    finally:
        signal.signal(sig, old)


def ignore_signal(sig):
    return set_handler(sig, signal.SIG_IGN)


def detach_ctty(ctty_fd):
    # TODO: will children receive sighup as well?
    # When a process detaches from a tty, it is sent the signals SIGHUP and then SIGCONT
    with ignore_signal(signal.SIGHUP):
        fcntl.ioctl(ctty_fd, termios.TIOCNOTTY)


def attach_ctty(fd: int) -> bool:
    try:
        fcntl.ioctl(fd, termios.TIOCSCTTY, 1)
        return True
    except PermissionError:
        return False


def get_ctty_fd() -> Optional[int]:
    """
    If there is a controlling tty for this process, return a read+write fd to it.
    Otherwise return None.
    """
    try:
        return os.open(os.ctermid(), os.O_RDWR)
    except OSError as e:
        if e.errno == errno.ENXIO:
            return None
        raise


def detach_current_ctty():
    """ Detach from ctty if there is one """
    ctty_fd = get_ctty_fd()
    # If there is no ctty, ctty_fd is None and we don't do anything
    if ctty_fd is not None:
        if is_session_leader():
            detach_ctty(ctty_fd)
        else:
            make_session_leader()
        os.close(ctty_fd)


@dataclass
class PTY:
    master_fd: int
    slave_fd: int
    _closed: bool = False

    def resize(self, rows, cols):
        winsize = struct.pack("HHHH", rows, cols, 0, 0)
        fcntl.ioctl(self.slave_fd, termios.TIOCSWINSZ, winsize)

    def close(self):
        if not self._closed:
            with ignore_signal(signal.SIGHUP):
                os.close(self.master_fd)
            self._closed = True

    def set_tty_attrs(self, tc_attrs, when=termios.TCSANOW):
        return
        IFLAG = 0
        OFLAG = 1
        CFLAG = 2
        LFLAG = 3
        ISPEED = 4
        OSPEED = 5
        CC = 6
        tc_attrs = tc_attrs[:]
        current_attrs = termios.tcgetattr(fd)
        tc_attrs[CC] = current_attrs[termios.CC]
        termios.tcsetattr(self.slave_fd, when, tc_attrs)

    def make_ctty(self) -> bool:
        detach_current_ctty()
        return attach_ctty(self.slave_fd)

    @classmethod
    @contextmanager
    def open(cls):
        master_fd, slave_fd = pty.openpty()
        self = cls(master_fd, slave_fd)
        try:
            yield self
        finally:
            self.close()


def print_to_ctty(string):
    """ If there is a ctty, print the given string to it. """
    ctty_fd = get_ctty_fd()
    if ctty_fd is not None:
        print(string, file=os.fdopen(ctty_fd, 'w'))



import pickle
import fcntl
import os
import struct
from collections import defaultdict
from functools import partial
from asyncio import new_event_loop
from io import BytesIO
from typing import Dict, Set

from .utils import opposite_dict

MESSAGE_LENGTH_FMT = 'I'


def set_nonblocking(fd):
    flags = fcntl.fcntl(fd, fcntl.F_GETFL)
    fcntl.fcntl(fd, fcntl.F_SETFL, flags | os.O_NONBLOCK)


def blocking_read(fd, n):
    io = BytesIO()
    read_amount = 0
    while read_amount < n:
        data = os.read(fd, n - read_amount)
        if not data:
            raise IOError('FD closed before all bytes read')
        read_amount += len(data)
        io.write(data)
    return io.getvalue()


class Piping:
    def __init__(self, pipe_dict: Dict[int, Set[int]]):
        self.buffers = defaultdict(bytes)
        self.loop = new_event_loop()
        for src_fd, dest_fds in pipe_dict.items():
            self.loop.add_reader(src_fd, partial(self._read, src_fd, dest_fds))
            for dest_fd in dest_fds:
                self.loop.add_writer(dest_fd, partial(self._write, dest_fd))
        self.readers_to_writers = dict(pipe_dict)
        self.writers_to_readers = opposite_dict(pipe_dict)

    def _remove_writer(self, writer_fd):
        self.loop.remove_writer(writer_fd)
        for reader_fd in self.writers_to_readers.pop(writer_fd):
            self.readers_to_writers.pop(reader_fd)

    def _remove_reader(self, reader_fd):
        # remove all writers that im the last to write to, remove all that write to me, if nothing left stop loop
        self.loop.remove_reader(reader_fd)
        writer_fds = self.readers_to_writers.pop(reader_fd)
        for writer_fd in writer_fds:
            writer_readers = self.writers_to_readers[writer_fd]
            writer_readers.remove(reader_fd)
            if not writer_readers:
                self._remove_writer(writer_fd)

    def _read(self, src_fd, dest_fds):
        try:
            data = os.read(src_fd, 1024)
        except OSError:
            data = ''
        if data:
            for dest_fd in dest_fds:
                self.buffers[dest_fd] += data
        else:
            self._remove_reader(src_fd)
            if src_fd in self.writers_to_readers:
                self._remove_writer(src_fd)
            if not self.readers_to_writers:
                self.loop.stop()

    def _write(self, dest_fd):
        buffer = self.buffers[dest_fd]
        if buffer:
            self.buffers[dest_fd] = buffer[os.write(dest_fd, buffer):]

    def run(self):
        self.loop.run_forever()
        # TODO: is this needed?
        # for dest_fd, buffer in self.buffers.items():
        #     while buffer:
        #         buffer = buffer[os.write(dest_fd, buffer):]


def send_message(sock, obj):
    message = pickle.dumps(obj)
    message_len = struct.pack(MESSAGE_LENGTH_FMT, len(message))
    sock.sendall(message_len)
    sock.sendall(message)


def receive_message(sock):
    len_len = struct.calcsize(MESSAGE_LENGTH_FMT)
    len_bytes = blocking_read(sock, len_len)
    message_len = struct.unpack(MESSAGE_LENGTH_FMT, len_bytes)[0]
    message = blocking_read(sock, message_len)
    return pickle.loads(message)



from __future__ import annotations
import runpy
import os
import socket
import sys
import traceback
from bdb import BdbQuit
from contextlib import contextmanager, nullcontext
from termios import tcdrain
from typing import Optional, ContextManager

from IPython.terminal.debugger import TerminalPdb
from IPython.terminal.interactiveshell import TerminalInteractiveShell
from prompt_toolkit.input.vt100 import Vt100Input
from prompt_toolkit.output.vt100 import Vt100_Output
from inspect import currentframe

from .utils import preserve_sys_state, run_thread
from .tty_utils import print_to_ctty, PTY
from .communication import receive_message, Piping


class RemoteIPythonDebugger(TerminalPdb):
    """
    Initializes IPython's TerminalPdb with stdio from a pty.
    As TerminalPdb uses prompt_toolkit instead of the builtin input(),
    we can use it to allow line editing and tab completion for files other than stdio (in this case, the pty).
    Because we need to provide the stdin and stdout params to the __init__, and they require a connection to the client,
    """
    _DEBUGGING_GLOBAL = 'DEBUGGING_WITH_MADBG'
    _CURRENT_INSTANCE = None

    @classmethod
    def _get_current_instance(cls) -> Optional[RemoteIPythonDebugger]:
        return cls._CURRENT_INSTANCE

    @classmethod
    def _set_current_instance(cls, new: Optional[RemoteIPythonDebugger]) -> None:
        cls._CURRENT_INSTANCE = new

    def __init__(self, stdin, stdout, term_type):
        # A patch until https://github.com/ipython/ipython/issues/11745 is solved
        TerminalInteractiveShell.simple_prompt = False
        term_input = Vt100Input(stdin)
        term_output = Vt100_Output.from_pty(stdout, term_type)
        super().__init__(pt_session_options=dict(input=term_input, output=term_output), stdin=stdin, stdout=stdout)
        self.use_rawinput = True
        self.done_callback = None

    def trace_dispatch(self, frame, event, arg, check_debugging_global=False):
        """
        Overriding super to allow the check_debugging_global and done_callback args.

        :param check_debugging_global: Whether to start debugging only if _DEBUGGING_GLOBAL is in the globals.
        :param done_callback: a callable to be called when the debug session ends.
        """
        if check_debugging_global:
            if self._DEBUGGING_GLOBAL in frame.f_globals:
                self.set_trace(frame)
            else:
                return None
        bdb_quit = False
        try:
            return super().trace_dispatch(frame, event, arg)
        except BdbQuit:
            bdb_quit = True
        finally:
            if self.quitting or bdb_quit:
                self._on_done()

    def _on_done(self):
        if self.done_callback is not None:
            self.done_callback()
            self.done_callback = None

    def set_trace(self, frame=None, done_callback=None):
        """ Overriding super to add the done_callback argument, allowing cleanup after a debug session """
        if done_callback is not None:
            # set_trace was called again without the previous one exiting -
            # happens on continue -> ctrl-c
            self.done_callback = done_callback
        if frame is None:
            frame = currentframe().f_back
        return super().set_trace(frame)

    def do_continue(self, arg):
        """ Overriding super to add a print """
        if not self.nosigint:
            print('Resuming program, press Ctrl-C to relaunch debugger.', file=self.stdout)
        return super().do_continue(arg)

    do_c = do_cont = do_continue

    def post_mortem(self, traceback):
        self.reset()
        self.interaction(None, traceback)

    def run_py(self, python_file, run_as_module, argv, set_trace=False):
        run_name = '__main__'
        globals = {self._DEBUGGING_GLOBAL: True}
        with preserve_sys_state():
            sys.argv = argv
            if not run_as_module:
                sys.path[0] = os.path.dirname(python_file)
            with self.debug(check_debugging_global=True) if set_trace else nullcontext():
                if run_as_module:
                    runpy.run_module(python_file, alter_sys=True, run_name=run_name, init_globals=globals)
                else:
                    runpy.run_path(python_file, run_name=run_name, init_globals=globals)

    @contextmanager
    def debug(self, check_debugging_global=False) -> ContextManager:
        self.reset()
        sys.settrace(lambda *args: self.trace_dispatch(*args, check_debugging_global=check_debugging_global))
        try:
            yield
        except BdbQuit:
            pass
        finally:
            self.quitting = True
            sys.settrace(None)

    @classmethod
    @contextmanager
    def start(cls, sock_fd: int) -> ContextManager[RemoteIPythonDebugger]:
        # TODO: just add to pipe list
        assert cls._get_current_instance() is None
        term_data = receive_message(sock_fd)
        term_attrs, term_type, term_size = term_data['term_attrs'], term_data['term_type'], term_data['term_size']
        with PTY.open() as pty:
            pty.resize(term_size[0], term_size[1])
            pty.set_tty_attrs(term_attrs)
            pty.make_ctty()
            piping = Piping({sock_fd: {pty.master_fd}, pty.master_fd: {sock_fd}})
            with run_thread(piping.run):
                slave_reader = os.fdopen(pty.slave_fd, 'r')
                slave_writer = os.fdopen(pty.slave_fd, 'w')
                try:
                    instance = cls(slave_reader, slave_writer, term_type)
                    cls._set_current_instance(instance)
                    yield instance
                except Exception:
                    print(traceback.format_exc(), file=slave_writer)
                    raise
                finally:
                    cls._set_current_instance(None)
                    print('Closing connection', file=slave_writer, flush=True)
                    tcdrain(pty.slave_fd)
                    slave_writer.close()

    @classmethod
    @contextmanager
    def get_server_socket(cls, ip: str, port: int) -> ContextManager[socket.socket]:
        """
        Return a new server socket for client to connect to. The caller is responsible for closing it.
        """
        server_socket = socket.socket()
        server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, True)
        server_socket.bind((ip, port))
        try:
            yield server_socket
        finally:
            server_socket.close()

    @classmethod
    @contextmanager
    def start_from_new_connection(cls, sock: socket.socket) -> ContextManager[RemoteIPythonDebugger]:
        print_to_ctty(f'Debugger client connected from {sock.getpeername()}')
        try:
            with cls.start(sock.fileno()) as debugger:
                yield debugger
        finally:
            sock.close()

    @classmethod
    def connect_and_start(cls, ip: str, port: int) -> ContextManager[RemoteIPythonDebugger]:
        # TODO: get rid of context managers at some level - nobody is going to use with start() anyway
        current_instance = cls._get_current_instance()
        if current_instance is not None:
            return nullcontext(current_instance)
        with cls.get_server_socket(ip, port) as server_socket:
            server_socket.listen(1)
            print_to_ctty(f'Waiting for debugger client on {ip}:{port}')
            sock, _ = server_socket.accept()
        return cls.start_from_new_connection(sock)



import sys
from click import ClickException, group, argument, option, pass_context

from madbg.client import connect_to_debugger
from madbg.consts import DEFAULT_IP, DEFAULT_PORT, DEFAULT_CONNECT_TIMEOUT
from madbg import run_with_debugging, attach_to_process

port_argument = argument('port', type=int, default=DEFAULT_PORT)
connect_timeout_option = option('-t', '--timeout', type=float, default=DEFAULT_CONNECT_TIMEOUT, show_default=True,
                                help='Connection timeout in seconds')


@group(context_settings=dict(help_option_names=['-h', '--help']))
def cli():
    pass


@cli.command()
@argument('ip', type=str, default=DEFAULT_IP)
@port_argument
@connect_timeout_option
def connect(ip, port, timeout):
    try:
        connect_to_debugger(ip, port, timeout=timeout)
    except (ConnectionRefusedError, TimeoutError):
        raise ClickException('Connection refused - did you use the right port?')


@cli.command()
@argument('pid', type=int)
@port_argument
@connect_timeout_option
def attach(pid, port, timeout):
    attach_to_process(pid, port, connect_timeout=timeout)


@cli.command(help='Run the given script or module with debugging features. '
                  'Flags given after the script name will be passed to the script as is.',
             context_settings=dict(ignore_unknown_options=True,
                                   allow_interspersed_args=False,
                                   allow_extra_args=True))
@option('-i', '--bind_ip', type=str, default=DEFAULT_IP, show_default=True)
@option('-p', '--port', type=int, default=DEFAULT_PORT, show_default=True)
@option('-n', '--no-post-mortem', is_flag=True, flag_value=True, default=False)
@option('-s', '--use-set-trace', is_flag=True, flag_value=True, default=False)
@option('-m', '--run-as-module', is_flag=True, flag_value=True, default=False, help='Works the same as python -m')
@argument('py_file', type=str, required=True)
@pass_context
def run(context, bind_ip, port, run_as_module, py_file, no_post_mortem, use_set_trace):
    argv = [sys.argv[0], *context.args]
    run_with_debugging(py_file, run_as_module=run_as_module, argv=argv, use_post_mortem=not no_post_mortem,
                       use_set_trace=use_set_trace, ip=bind_ip, port=port)


if __name__ == '__main__':
    cli()



import os
import socket
import time
import atexit
from functools import partial
from tty import setraw
from termios import tcdrain, tcgetattr, tcsetattr, TCSANOW
from contextlib import contextmanager

from .communication import Piping, send_message
from .consts import DEFAULT_IP, DEFAULT_PORT, STDIN_FILENO, STDOUT_FILENO, DEFAULT_CONNECT_TIMEOUT


def get_tty_handle():
    return os.open(os.ctermid(), os.O_RDWR)


@contextmanager
def tmp_atexit(func, *args, **kwargs):
    atexit.register(func, *args, **kwargs)
    try:
        yield
    finally:
        atexit.unregister(func)


@contextmanager
def promise_cleanup(func, cleanup):
    with tmp_atexit(cleanup):
        try:
            yield func()
        finally:
            cleanup()


def prepare_terminal():
    tty_handle = get_tty_handle()
    old_tty_mode = tcgetattr(tty_handle)
    set_raw = partial(setraw, tty_handle, TCSANOW)
    cleanup = partial(tcsetattr, tty_handle, TCSANOW, old_tty_mode)
    return promise_cleanup(set_raw, cleanup)


@contextmanager
def connect_to_server(ip, port, timeout):
    original_timeout = timeout
    start_time = time.time()
    s = None
    while not s:
        try:
            s = socket.create_connection((ip, port), timeout=timeout)
        except ConnectionRefusedError:
            pass
        timeout = original_timeout - (time.time() - start_time)
        if timeout <= 0:
            raise TimeoutError()
    try:
        yield s
    finally:
        s.close()


def connect_to_debugger(ip=DEFAULT_IP, port=DEFAULT_PORT, timeout=DEFAULT_CONNECT_TIMEOUT,
                        in_fd=STDIN_FILENO, out_fd=STDOUT_FILENO):
    with connect_to_server(ip, port, timeout) as socket:
        tty_handle = get_tty_handle()
        term_size = os.get_terminal_size(tty_handle)
        term_data = dict(term_attrs=tcgetattr(tty_handle),
                         # prompt toolkit will receive this string, and it can be 'unknown'
                         term_type=os.environ.get("TERM", "unknown"),
                         term_size=(term_size.lines, term_size.columns))
        send_message(socket, term_data)
        with prepare_terminal():
            socket_fd = socket.fileno()
            Piping({in_fd: {socket_fd}, socket_fd: {out_fd}}).run()
            tcdrain(out_fd)



import os
import re
import signal
import sys
from select import select
from traceback import format_exc
from contextlib import nullcontext
from inspect import currentframe
from pdb import Restart
from hypno import inject_py
from fcntl import fcntl, F_GETFL, F_SETFL, F_SETOWN
from os import O_ASYNC, getpid

from .client import connect_to_debugger
from .tty_utils import print_to_ctty, set_handler
from .utils import use_context
from .consts import DEFAULT_IP, DEFAULT_PORT, DEFAULT_CONNECT_TIMEOUT
from .debugger import RemoteIPythonDebugger

DEBUGGER_CONNECTED_SIGNAL = signal.SIGUSR1


def _inject_set_trace(pid, ip=DEFAULT_IP, port=DEFAULT_PORT):
    assert isinstance(ip, str)
    assert re.fullmatch('[.0-9]+', ip)
    assert isinstance(port, int)
    sig_num = DEBUGGER_CONNECTED_SIGNAL.value
    inject_py(pid, f'__import__("signal").signal({sig_num},lambda _,f:__import__("madbg").set_trace(f,"{ip}",{port}))')
    os.kill(pid, sig_num)


def attach_to_process(pid: int, port=DEFAULT_PORT, connect_timeout=DEFAULT_CONNECT_TIMEOUT):
    ip = '127.0.0.1'
    _inject_set_trace(pid, ip, port)
    connect_to_debugger(ip, port, timeout=connect_timeout)


def set_trace(frame=None, ip=DEFAULT_IP, port=DEFAULT_PORT):
    if frame is None:
        frame = currentframe().f_back
    debugger, exit_stack = use_context(RemoteIPythonDebugger.connect_and_start(ip, port))
    debugger.set_trace(frame, done_callback=exit_stack.close)


def set_trace_on_connect(ip=DEFAULT_IP, port=DEFAULT_PORT):
    """
    Set up a debugger in another thread, which will signal the main thread when it receives a connection.
    Also set up a signal handler that will call set_trace when the signal is received.
    """
    server_socket, server_exit_stack = use_context(RemoteIPythonDebugger.get_server_socket(ip, port))

    def sigio_handler(signum, frame):
        if select([server_socket], [], [], 0)[0]:
            handler_exit_stack.close()
            sock, _ = server_socket.accept()
            server_exit_stack.close()
            debugger, debugger_exit_stack = use_context(RemoteIPythonDebugger.start_from_new_connection(sock))

            def on_trace_done():
                debugger_exit_stack.close()
                set_trace_on_connect(ip, port)

            debugger.set_trace(frame, done_callback=on_trace_done)
        elif not isinstance(old_handler, signal.Handlers):
            old_handler(signum, frame)

    old_handler, handler_exit_stack = use_context(set_handler(signal.SIGIO, sigio_handler))
    server_fd = server_socket.fileno()
    fcntl(server_fd, F_SETOWN, getpid())
    fcntl(server_fd, F_SETFL, fcntl(server_fd, F_GETFL, 0) | O_ASYNC)
    server_socket.listen(1)
    print_to_ctty(f'Listening for debugger client on {ip}:{port}')


def post_mortem(traceback=None, ip=DEFAULT_IP, port=DEFAULT_PORT):
    traceback = traceback or sys.exc_info()[2] or sys.last_traceback
    with RemoteIPythonDebugger.connect_and_start(ip, port) as debugger:
        debugger.post_mortem(traceback)


def run_with_debugging(python_file, run_as_module=False, argv=(), use_post_mortem=True, use_set_trace=False,
                       ip=DEFAULT_IP, port=DEFAULT_PORT, debugger=None):
    argv = [python_file, *argv]
    with RemoteIPythonDebugger.connect_and_start(ip, port) if debugger is None else nullcontext(debugger) as debugger:
        try:
            debugger.run_py(python_file, run_as_module, argv, set_trace=use_set_trace)
        except Restart:
            print("Restarting", python_file, "with arguments:", file=debugger.stdout)
            print("\t" + " ".join(argv), file=debugger.stdout)
            return run_with_debugging(python_file, run_as_module=run_as_module, argv=argv,
                                      use_post_mortem=use_post_mortem, use_set_trace=use_set_trace,
                                      ip=ip, port=port, debugger=debugger)
        except SystemExit as e:
            print(f"The program exited via sys.exit(). Exit status: {e.code}", end=' ', file=debugger.stdout)
        except SyntaxError:
            raise
        except:
            if use_post_mortem:
                print(format_exc(), file=debugger.stdout)
                debugger.post_mortem(sys.exc_info()[2])
            raise
        else:
            print(f'{python_file} finished running successfully', file=debugger.stdout)


__all__ = ['attach_to_process', 'set_trace', 'set_trace_on_connect', 'post_mortem', 'run_with_debugging']



import atexit
import sys
import threading
from collections import defaultdict
from concurrent.futures.thread import ThreadPoolExecutor
from contextlib import contextmanager, ExitStack
from typing import Dict, Any, Set


@contextmanager
def preserve_sys_state():
    sys_argv = sys.argv[:]
    sys_path = sys.path[:]
    try:
        yield
    finally:
        sys.argv = sys_argv
        sys.path = sys_path


def register_atexit(callback, *args, **kwargs):
    if sys.version_info >= (3, 9):
        # Since python3.9, ThreadPoolExecutor threads are non-daemon, which means they are joined before atexit
        # hooks run - https://bugs.python.org/issue39812
        threading._register_atexit(callback, *args, **kwargs)
    else:
        atexit.register(callback, *args, **kwargs)


def use_context(context_manager, exit_stack=None):
    if exit_stack is None:
        exit_stack = ExitStack()
        register_atexit(exit_stack.close)
    context_value = exit_stack.enter_context(context_manager)
    return context_value, exit_stack


@contextmanager
def run_thread(func, *args, **kwargs):
    with ThreadPoolExecutor(1) as executor:
        future = executor.submit(func, *args, **kwargs)
        try:
            yield future
        finally:
            future.result()


def opposite_dict(dict_: Dict[Any, Set[Any]]) -> Dict[Any, Set[Any]]:
    opposite = defaultdict(set)
    for key, values in dict_.items():
        for value in values:
            opposite[value].add(key)
    return opposite
